#include <bpkgdb/create.h>

#include <bpkgdb/sha512.h>
#include <bpkgdb/util.h>

#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>

extern int argc;
extern char** argv;
extern char* config_file;
extern char* output;

struct package {
	char* name;
	char* shortdesc;
	char* version;
	char* dependencies;
};

int bpkgdb_create(){
	if(config_file == NULL){
		fprintf(stderr, "ERR : Config file was not specified.\n");
		return 1;
	}else{
		if(access(config_file, R_OK) == 0){
			int counter = 0;
			char* buffer = malloc(1);
			buffer[0] = 0;
			char tmpbuf;
			FILE* f = fopen(config_file, "r");
			struct package** packages = malloc(sizeof(*packages));
			int package_count = 0;
			packages[0] = NULL;
			char* name = malloc(1);
			char* shortdesc = malloc(1);
			char* version = malloc(1);
			char* dbname = sstrdup("BPKG Database");
			char* dbdest = sstrdup("bpkg.db");
			char* dependencies = malloc(1);
			remove(dbdest);
			name[0] = 0;
			shortdesc[0] = 0;
			version[0] = 0;
			dependencies[0] = 0;
			while(1){
				fread(&tmpbuf, 1, 1, f);
				if(tmpbuf == '\n' || feof(f)){
					if(counter != 0){
						if(buffer[0] == '[' && buffer[strlen(buffer) - 1] == ']'){
							char* prop = malloc(strlen(buffer) - 2 + 1);
							prop[strlen(buffer) - 2] = 0;
							memcpy(prop, buffer + 1, strlen(buffer) - 2);
							if(strcmp(prop, "/") == 0){
								packages = realloc(packages, sizeof(*packages) * (package_count + 2));
								struct package* pkg = malloc(sizeof(*pkg));
								pkg->name = sstrdup(name);
								pkg->shortdesc = sstrdup(shortdesc);
								pkg->version = sstrdup(version);
								pkg->dependencies = sstrdup(dependencies);
								packages[package_count] = pkg;
								packages[package_count + 1] = NULL;
								package_count++;
								free(name);
								free(shortdesc);
								free(version);
								free(dependencies);
								name = malloc(1);
								shortdesc = malloc(1);
								version = malloc(1);
								dependencies = malloc(1);
								name[0] = 0;
								shortdesc[0] = 0;
								version[0] = 0;
								dependencies[0] = 0;
							}else{
								free(name);
								name = sstrdup(prop);
							}
							free(prop);
						}else{
							char* key = buffer;
							int i;
							for(i = 0; buffer[i] != 0 && buffer[i] != '='; i++);
							key[i] = 0;
							char* value = buffer + i + 1;
							if(strcmp(key, "Version") == 0){
								free(version);
								version = sstrdup(value);
							}else if(strcmp(key, "ShortDescription") == 0){
								free(shortdesc);
								shortdesc = sstrdup(value);
							}else if(strcmp(key, "Dependencies") == 0){
								int i = 0;
								char* str = malloc(1);
								str[0] = 0;
								int strcounter = 0;
								while(1){
									char ch = value[i];
									if(ch == 0 || ch == ' '){
										if(strlen(dependencies) != 0){
											char* tmp = dependencies;
											dependencies = sstrcat(tmp, "\n");
											free(tmp);
										}
										char* tmp = dependencies;
										dependencies = sstrcat(tmp, str);
										free(tmp);
										strcounter = 0;
										if(ch == 0) break;
									}else{
										str = realloc(str, strcounter + 2);
										str[strcounter] = ch;
										str[strcounter + 1] = 0;
										strcounter++;
									}
									i++;
								}
								free(str);
							}else if(strcmp(key, "DatabaseName") == 0){
								free(dbname);
								dbname = sstrdup(value);
							}

						}
					}
					counter = 0;
					free(buffer);
					buffer = malloc(1);
					buffer[0] = 0;
					if(feof(f)) break;
				}else{
					buffer = realloc(buffer, counter + 2);
					buffer[counter] = tmpbuf;
					buffer[counter + 1] = 0;
					counter++;
				}
			}
			printf("Database name is `%s`, length %d\n", dbname, strlen(dbname));
			uint16_t dbnamelen = strlen(dbname);
			FILE* db;
			if((db = fopen(dbdest, "w")) == NULL){
				fprintf(stderr, "ERR : Couldn't open the database file!\n");
				return 1;
			}
			printf("Writing the signature ... ");
			fflush(stdout);
			fwrite("\177BPKG", 5, 1, db);
			printf("done.\n");
			printf("Writing the database description ... ");
			fflush(stdout);
			uint8_t ch;
			ch = dbnamelen & 0xff;
			fwrite(&ch, 1, 1, db);
			ch = dbnamelen >> 8;
			fwrite(&ch, 1, 1, db);
			fwrite(dbname, strlen(dbname), 1, db);
			printf("done.\n");
			printf("Writing the package information.\n");
			char* state = sstrdup("");
			int i;
			for(counter = 0; packages[counter] != NULL; counter++);
			for(i = 0; packages[i] != NULL; i++){
				struct package* pkg = packages[i];
				char* fullname = sstrcat3(pkg->name, "-", pkg->version);
				printf("\r[%6d/%6d | 1/3]\x1b[0K %16s : Creating the tarball", i + 1, counter, pkg->name);
				fflush(stdout);
				pid_t pid = fork();
				if(pid == 0){
					chdir(pkg->name);
					remove(sstrcat3("../", fullname, ".tar.gz"));
					execlp("tar", "tar", "czf", sstrcat3("../", fullname, ".tar.gz"), ".", NULL);
					_exit(0);
				}else{
					waitpid(pid, 0, 0);
				}
				printf("\r[%6d/%6d | 2/3]\x1b[0K %16s : Writing the entry", i + 1, counter, pkg->name);
				fflush(stdout);
				fprintf(db, "%.3d", strlen(pkg->name));
				fprintf(db, "%s", pkg->name);
				fprintf(db, "%.2d", strlen(pkg->version));
				fprintf(db, "%s", pkg->version);
				fprintf(db, "%.3d", strlen(pkg->shortdesc));
				fprintf(db, "%s", pkg->shortdesc);
				fprintf(db, "%.16d", strlen(pkg->dependencies));
				fprintf(db, "%s", pkg->dependencies);
				printf("\r[%6d/%6d | 3/3]\x1b[0K %16s : Writing the SHA512", i + 1, counter, pkg->name);
				fflush(stdout);
				char* path = sstrcat(fullname, ".tar.gz");
				FILE* ftarball = fopen(path, "r");
				if(ftarball == NULL){
					free(state);
					free(path);
					state = sstrdup(" (not correctly)");
					printf("%s", state);
					fflush(stdout);
					break;
				}else{
					struct stat s;
					stat(path, &s);
					char* content = malloc(s.st_size);
					fread(content, s.st_size, 1, ftarball);
					fclose(ftarball);
					free(path);
					uint8_t* sha512 = bpkgdb_sha512(content, s.st_size);
					free(content);
					char* dsha512 = malloc(129);
					memset(dsha512, 0, 129);
					int j;
					const char* hex = "0123456789abcdef";
					for(j = 0; j < 64; j++){
						dsha512[j * 2] = hex[(sha512[j] & 0xf0) >> 4];
						dsha512[j * 2 + 1] = hex[sha512[j] & 0xf];
					}
					fprintf(db, "%s", dsha512);
					free(dsha512);
					free(sha512);
				}
			}
			fprintf(db, "%.3d", 0);
			fclose(db);
			printf("\n");
			free(name);
			free(version);
			free(shortdesc);
			free(dbname);
			fclose(f);
			free(buffer);
			for(i = 0; packages[i] != NULL; i++){
				free(packages[i]->name);
				free(packages[i]->version);
				free(packages[i]->shortdesc);
				free(packages[i]->dependencies);
				free(packages[i]);
			}
			free(packages);
			printf("Finished%s.\n", state);
			return 0;
		}else{
			fprintf(stderr, "ERR : Config file was not readable or non-existent.\n");
			return 1;
		}
	}
}
